#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <poll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "picohttpparser.h"

#include "event_loop.h"
#include "http.h"

void event_loop_conn_io_res(event_loop_conn *conn) {
  while (1) {
    ssize_t res = 0;
    size_t remain = conn->wbuf_size - conn->wbuf_sent;

    do {
      res = write(conn->fd, &conn->wbuf[conn->wbuf_sent], remain);
    } while (res < 0 && errno == EINTR);

    // EAGAIN doesn't mean there was an error, but rather that there's no more
    // data right now, but there might be more later, aka "try again later"
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // If it's not EGAIN, there was an error writing so we simply end the
    // request
    if (res < 0) {
      conn->state = event_loop_conn_state_end;
      return;
    }

    conn->wbuf_sent += (size_t)res;

    // After writing a response, we switch back to request mode to process a
    // next request
    if (conn->wbuf_sent == conn->wbuf_size) {
      conn->state = conn->close_after_write ? event_loop_conn_state_end
                                            : event_loop_conn_state_req;
      /* c->wbuf_sent = 0; */
      /* c->wbuf_size = 0; */

      return;
    }
  }
}

bool event_loop_handle_request(event_loop_conn *conn) {
  // Prevents the request handler function from looping indefinitely without
  // ever consuming new data
  if (conn->rbuf_size - conn->rbuf_read == 0) {
    return false;
  }

  http_parse_error res = http_parse_request(
      &conn->ctx->req, (const char *)&conn->rbuf[conn->rbuf_read],
      conn->rbuf_size - conn->rbuf_read);

  if (res == http_parse_error_ok) {
    // Perform the request
    http_route(conn);
  }
  // Both in the case of an invalid HTTP request or one that's larger than the
  // read buffer, we cannot determine when the next, possibly valid, HTTP
  // request begins in the data stream. Therefore, we close the connection,
  // even if additional pipelined requests are incoming.
  else if (res == http_parse_error_invalid ||
           (res == http_parse_error_incomplete &&
            conn->rbuf_size == EVENT_LOOP_BUFFER_SIZE)) {
    conn->state = event_loop_conn_state_end;
  }

  // TODO in highly concurrent situations, it might actually be better to always
  // return false here, as this allows cycling better through all connections
  return conn->state == event_loop_conn_state_req;
}

/**
 * Read new data into the read buffer. This command performs at most one
 * successful read syscall.
 *
 * Returns whether the function should be retried immediately or not.
 */
void event_loop_conn_io_req(event_loop_conn *conn) {
  do {
    // Move remaining data to start of buffer
    memmove(conn->rbuf, &conn->rbuf[conn->rbuf_read],
            conn->rbuf_size - conn->rbuf_read);
    conn->rbuf_size -= conn->rbuf_read;
    conn->rbuf_read = 0;

    ssize_t res;
    size_t cap = EVENT_LOOP_BUFFER_SIZE - conn->rbuf_size;

    // Try to read at most cap bytes from the file descriptor
    do {
      res = read(conn->fd, &conn->rbuf[conn->rbuf_size], cap);
    } while (res < 0 && errno == EINTR);

    // EGAIN means we try again later
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // Any other negative error message means the read errored out
    if (res < 0) {
      conn->state = event_loop_conn_state_end;

      return;
    }

    // Output of 0 and no error means we've reached the end of the input. We run
    // try_one_request one more time and close the connection if this doesn't
    // change the result.
    if (res == 0) {
      conn->state = event_loop_conn_state_end;

      return;
    }

    // We switch to processing mode if we've reached the end of the data stream,
    // or if the read buffer is filled
    /* if (res == 0 || c->rbuf_size == MAX_MSG_SIZE) { */
    /*     c->state = STATE_PROCESS; */
    /*     return false; */
    /* } */

    conn->rbuf_size += (size_t)res;

    // This loop allows processing multiple requests from a single read buffer
    while (event_loop_handle_request(conn))
      ;
  }
  // We can keep reading as long as we're in request mode
  while (conn->state == event_loop_conn_state_req);
}

void event_loop_conn_io(event_loop_conn *conn) {
  switch (conn->state) {
  case event_loop_conn_state_req:
    event_loop_conn_io_req(conn);
    break;
  case event_loop_conn_state_res:
    event_loop_conn_io_res(conn);
    break;
  case event_loop_conn_state_end:
    printf("we shouldn't be here\n");
    break;
  }
}
